/* ========================================
   Connect Four Robot AI Software
    Written By: James McCabe
    6.115 Final Project, 05/14/14
    
    Appendix B
 * ========================================
*/
#include <project.h>
#include <Math.h>

int score(int board[6][7]);
void initializeBoard();
int  negamax(int board[6][7], int parent, int topOfColumn[7], int depth, int color);
void makeMove(int board[6][7], int column, int player, int topOfColumn[7]);
void undoMove(int board[6][7], int column, int topOfColumn[]);
int getUserMove();
int hint();
int getDepth();
char getUserAction();
int getUserParam();
void waitForAccept();

// stores the state of the game board
int currentBoard[6][7];
// stores the heights of each column
int currentTopOfColumn[] = {-1,-1,-1,-1,-1,-1,-1};
// the maximum value returned by the heuristic funtion
int maxVal = 500000;
// the difficulty setting of the robot
int difficulty = 1;
// indicates of the user has made their first move or not
int madeMove = -1;


int main(){   
    LCD_Char_1_Start();	            // initialize lcd
	LCD_Char_1_ClearDisplay();      // clear the lcd
    UART_Start();                   // initialize the UART comm.
    initializeBoard();              // initialize the internal board
    
    while(1){
        int userMove =  - 1;                // stores the user's move
        char userAction = getUserAction();  // wait for and get the action from the user
        
        if(userAction == 'H'){              // the user wants a hint
            int hintForUser = hint()+1;     // calculate the hint
			
            UART_PutChar('H');              // respond first with the hint action character
			waitForAccept();                // wait for the 2051 to ask for the parameter
            CyDelay(100);                   
            UART_PutChar(hintForUser);      // send the actual hint to the 2051
            
			// display that hint on the LCD for debugging purposes                            
            LCD_Char_1_ClearDisplay();
            LCD_Char_1_Position(0, 0);
	        LCD_Char_1_PrintString("hint: ");		
	        LCD_Char_1_PrintNumber(hintForUser);
        } else if(userAction == 'D' && madeMove == -1){ // the user has set the difficulty
            difficulty = getUserParam();       // the difficulty is the parameter sent by the 2051
            
            // display the difficulty on the LCD for debugging
            LCD_Char_1_ClearDisplay();
            LCD_Char_1_Position(0, 0);
	        LCD_Char_1_PrintString("difficulty: ");		
	        LCD_Char_1_PrintNumber(difficulty);
        }  else if(userAction == 'M'){  // the user has made a move
            userMove = getUserParam();  // the user's move is the parameter sent by the 2051
        }
        
        UART_ClearRxBuffer();
        UART_ClearTxBuffer();
        
        // the robot will only make a move after the user has made one
        if(userMove != -1){
            // set the indicator variable that indicates the user has made their first move
            // this prevents them from being able to change the difficulty mid game
            madeMove = 1;
            // display the user's move on the LCD for debugging
            LCD_Char_1_ClearDisplay();
            LCD_Char_1_Position(0, 0);
	        LCD_Char_1_PrintString("your move: ");		
	        LCD_Char_1_PrintNumber(userMove);
            
            // make the user's move on the internal game board representation
            makeMove(currentBoard, userMove-1, 1, currentTopOfColumn);
            
            // if the user has won then let them know that
            if(score(currentBoard) == -maxVal){
                // display 'you win' on LCD for debugging purposes
                LCD_Char_1_ClearDisplay();
                LCD_Char_1_Position(0, 0);
                LCD_Char_1_PrintString("you win!");
                // send the "user won" action character to the 2051
                UART_PutChar('W');
                // exit the program loop and enter a do nothing loop
                break;
            }
            
            // calculate the robot's move using the Negamax algorithm
            int robotMove = negamax(currentBoard, -1, currentTopOfColumn, getDepth(), 1);
            // make the robot's move on the internal game board representation
            makeMove(currentBoard, robotMove, 2, currentTopOfColumn);
            // display the robot's move on the LCD for debugging
            LCD_Char_1_Position(1, 0);
	        LCD_Char_1_PrintString("robot move: ");
	        LCD_Char_1_PrintNumber(robotMove+1);
            
            // if the robot will win after it makes its move then let the
            // 2051 know that by sending the 'you lose' action character
            // before sending the robot's move.
            if(score(currentBoard) == maxVal){
                LCD_Char_1_Position(0, 0);
                LCD_Char_1_PrintString("you lose!!!!!");
				
                // send the 'you lose' action character to the 2051
                UART_PutChar('L');
                // wait for the 2051 to ask for the parameter
                waitForAccept();
            	CyDelay(100);
                // send the robot's move to the 2051
                UART_PutChar(robotMove+1);
				// exit the program loop and enter a do nothing loop
                break;
            }
            
            //send the normal 'move' action character
            UART_PutChar('M');
            // wait for the 2051 to ask for the parameter
            waitForAccept();
            CyDelay(100);
            // send the robot's move to the 2051
            UART_PutChar(robotMove+1);
        }
    }
    
    // the do nothing loop that is entered when someone wins
    while(1){
    }
}


/*
Blocks, waiting to receive an action character from the 2051.
The action character represents what type of action the user has made.
This can be asking for a hint ('H'), setting the difficulty ('D'),
or making a move ('M').
*/
char getUserAction(){
    // wait until there is a byte in the buffer
    while(UART_GetRxBufferSize() == 0){
        CyDelay(100);
    }
    // return the received action character
    return (char)UART_GetChar();
}


/*
Blocks, waiting to receive the parameter from the 2051 that
follows an action character. Only actions 'D' and 'M' have
parameters. Asking for a hint, 'H', requires no additional
parameter.
*/
int getUserParam(){
    // wait until there is a byte in the buffer
    while(UART_GetRxBufferSize() == 0){
        CyDelay(100);
    }
    // return the received parameter
    return UART_GetChar();
}


/*
Blocks, waiting for the 2051 to ask for the parameter
that follows an action character sent by the PSOC.
*/
void waitForAccept(){
	char result = 'a';
	while(result != 'P'){
		while(UART_GetRxBufferSize() == 0){
        	CyDelay(100);
    	}
		result = (char)UART_GetChar();
	}
}


/**
Returns a hint for which move the user should make. The hint
is an number between 0-6, representing a column on the game board
from left to right.
**/
int hint(){
    // calculate the depth that should be used to do the search
    int depth = getDepth();
    // run negamax, but as the user and not as the robot
    return negamax(currentBoard, -1, currentTopOfColumn, depth, -1);
}


/*
Sets all of the initial locations in the board to 0 - meaning
they are empty.
*/
void initializeBoard(){
    int row;
    for (row = 0; row < 6; row++){          // iterate through all of the rows
        int col;
        for (col = 0; col < 7; col++){      // iterate through all of the columns
            currentBoard[row][col] = 0;     // set each element to 0 (empty)
        }
    }
}


/*
Calculates the depth that should be used when during the search in the
Negamax algorithm. The depth depends on the difficulty and the number
of possible moves. When the number of possible moves decreases (i.e. 
columns fill up) the algorithm can search deeper in the same amount
of time.
*/
int getDepth(){
    int numPossibleMoves = 0;
    
    // iterate through the columns and count how many are not full
    // this gives the number of possible moves that can currently be made
    int col;
    for(col = 0; col < 7; col++){
        numPossibleMoves += (currentTopOfColumn[col] < 5) ? 1 : 0;
    }
    
    // based off of the number of possible moves and the difficulty,
    // calculate and return the search depth.
    if(numPossibleMoves ==  4){
        return 5 - (4-difficulty);
    } else if(numPossibleMoves == 3 || numPossibleMoves == 1){
        return 7 - (4-difficulty);
    } else if(numPossibleMoves == 2){
        return 11 - (4-difficulty);
    } else{
        return 4 - (4-difficulty);
    }
}



/*
Runs the Negamax algorithm (en.wikipedia.org/wiki/Negamax) on the game board to 
find the best move to make for a given player. 'Best' is defined by the heuristic 
funtion that is used. The returned value is a number between 0-6, representing a column
on the game board from left to right.
*/
int negamax(int board[6][7], int parent, int topOfColumn[7], int depth, int color){
    // run the heuristic function on the board to calculate the score, which is in terms
    // of the robot. 
    int scoreOfBoard = score(board);
    
    // base case is at depth = 0 or someone has won
    if (depth == 0 || scoreOfBoard == maxVal || scoreOfBoard == -maxVal){
        // before returning we must undo the hypothetical move that was made
        undoMove(board, parent, topOfColumn);
        // a positive score is good for the robot, a negative score is good for the user,
        // which means when maximizing the scores we multiply by -1 for the user.
        return (color * scoreOfBoard);
    }
    
    int bestValue = -5000000;
    int bestChild = 0 ;
    int hasChildren = -1;

    int col;
    for (col = 0; col < 7; col++){
        // if the column is not full then there are children to explore
        if(topOfColumn[col] < 5){ 
            hasChildren = 1;
            int child = col;
            
            // the robot's pieces are represent as '2' and the user's as '1'
            int player = (color > 0) ? 2 : 1;
            // make the hypothetical move on the game board
            makeMove(board, child, player, topOfColumn);
            // recursively run Negamax on the resulting board - switching whose
            // turn it is.
            int val = -negamax(board, child, topOfColumn, depth-1, -color);
            // keeps track of the best value and best child (best move to make)
            if(val >= bestValue){
                bestValue  = val;
                bestChild = child;
            }
        }
    }
    
    // return the best value that was found
    // if this is not the root node
    if(parent != -1 && hasChildren == 1){
        undoMove(board,parent, topOfColumn);
        return bestValue;
    } else if(parent != -1 && hasChildren == -1){
        undoMove(board, parent, topOfColumn);
        return (color * scoreOfBoard);
    }
    // the root node returns the best child (best move)
    else{
        return bestChild;
    }
}


/*
Makes a given move for a given player on the given board
*/
void makeMove(int board[6][7], int column, int player, int topOfColumn[7]){
    int piece = player; // 1 is the human, 2 is the robot
    if(topOfColumn[column] < 5){
        topOfColumn[column]++;
        board[topOfColumn[column]][column] = piece;
    }
}



/*
Undo the last move on the given column on the given board
*/
void undoMove(int board[6][7], int column, int topOfColumn[]){
    // set the location of the last move to 0 (empty)
    board[topOfColumn[column]][column] = 0;
    // shorten the height of that column
    topOfColumn[column]--;
}




/*
Scores the current board using #-out-4 heuristic
*/
int score(int board[6][7]){
    int robotLineCount[] = {0,0,0,0,0};
    int humanLineCount[] = {0,0,0,0,0};
    
    // score columns
    
    int col;
    for (col = 0; col < 7; col++){           // iterate through all of the columns
        int base;
        for(base = 5; base >= 3; base--){   // iterate through all the possible bases for lines of 4
            int piecesInLineRobot = 0;
            int piecesInLineHuman = 0;
            // iterates along a line
            int offset;
            for(offset = 0; offset < 4; offset++){
                int row = base - offset;
                piecesInLineRobot += (board[row][col] == 2) ? 1 : 0;
                piecesInLineHuman += (board[row][col] == 1) ? 1 : 0;
            }
            // scores only count if a line does not have both human and robot pieces in it
            if(!(piecesInLineRobot > 0 && piecesInLineHuman > 0)){
                robotLineCount[piecesInLineRobot]++;
                humanLineCount[piecesInLineHuman]++;
            }
        }
    }
    
    // score rows
    int row;
    for (row = 0; row < 6; row++){
        int base;
        for (base = 0; base <= 3; base++){
            int piecesInLineRobot = 0;
            int piecesInLineHuman = 0;
            
            // iterates along a line
            int offset;
            for (offset = 0; offset < 4; offset++){
                int col = base + offset;
                piecesInLineRobot += (board[row][col] == 2) ? 1 : 0;
                piecesInLineHuman += (board[row][col] == 1) ? 1 : 0;
            }
            // scores only count if a line does not have both human and robot pieces in it
            if(!(piecesInLineRobot > 0 && piecesInLineHuman > 0)){
                robotLineCount[piecesInLineRobot]++;
                humanLineCount[piecesInLineHuman]++;
            }  
        }
    }
    
    // score diagonals left-to-right
    int rowBases1[] = {3,4,5,5,5,5};
    int colBases1[] = {0,0,0,1,2,3};
    int numLines[]  = {1,2,3,3,2,1};
    int numDiagonals = 6;
    
    int diagonal;
    for (diagonal = 0; diagonal < numDiagonals; diagonal++){
        int base;
        for (base = 0; base < numLines[diagonal]; base++){
            int piecesInLineRobot = 0;
            int piecesInLineHuman = 0;
            
            int offset;
            for(offset = 0; offset < 4; offset++){
                int col = colBases1[diagonal] + base + offset;
                int row = rowBases1[diagonal] - base - offset;
                piecesInLineRobot += (board[row][col] == 2) ? 1 : 0;
                piecesInLineHuman += (board[row][col] == 1) ? 1 : 0;
            }
            // scores only count if a line does not have both human and robot pieces in it
            if(!(piecesInLineRobot > 0 && piecesInLineHuman > 0)){
                robotLineCount[piecesInLineRobot]++;
                humanLineCount[piecesInLineHuman]++;
            }  
        }
    }
    
    // score diagonals right-to-left
    int rowBases2[] = {3,4,5,5,5,5};
    int colBases2[] = {6,6,6,5,4,3};
    
    for (diagonal = 0; diagonal < numDiagonals; diagonal++){
        int base;
        for (base = 0; base < numLines[diagonal]; base++){
            int piecesInLineRobot = 0;
            int piecesInLineHuman = 0;
            
            int offset;
            for(offset = 0; offset < 4; offset++){
                int col = colBases2[diagonal] - base - offset;
                int row = rowBases2[diagonal] - base - offset;
                piecesInLineRobot += (board[row][col] == 2) ? 1 : 0;
                piecesInLineHuman += (board[row][col] == 1) ? 1 : 0;
            }
            // scores only count if a line does not have both human and robot pieces in it
            if(!(piecesInLineRobot > 0 && piecesInLineHuman > 0)){
                robotLineCount[piecesInLineRobot]++;
                humanLineCount[piecesInLineHuman]++;
            }  
        }
    }
    
    // the scores are the weighted sum of each count
    int robotScore = robotLineCount[1] + robotLineCount[2]*20 + robotLineCount[3]*100;
    int humanScore = humanLineCount[1] + humanLineCount[2]*20 + humanLineCount[3]*100;
    
    // if robot has won then the score is the max value
    if(robotLineCount[4] > 0){
        robotScore = maxVal;
        humanScore = 0;
    }
    // if the human has won then the score is the min value
    if(humanLineCount[4] > 0){
        robotScore = 0;
        humanScore = maxVal;
    }
    // the score is in terms of the robot, so it is the difference
    // between the robot and human score.
    int score = robotScore - humanScore;  
    return score;
}


/* [] END OF FILE */
